local mStore = require "mStore";

local mModule = {
    new = function(cls,name)
        local mod = setmetatable({},cls:mt());
        mod.name = name;
        mod.subModules = {};
        return mod;
    end,
    prototype = {
        addModule = function(self,mod)
            print(string.format("Adding submodule %s to %s",mod.name,self.name));
            self.subModules[mod.name] = mod;
        end,
        init = function(self)
            for i,v in pairs(self.subModules) do
                v:init();
            end    
        end,
        Start = function(self)
            for i,v in pairs(self.subModules) do
                v:Start();
            end
        end,
        GameKey = function(self,key)
            for i,v in pairs(self.subModules) do
                v:GameKey(key);
            end
        end,
        AddObject = function(self,handle)
            for i,v in pairs(self.subModules) do
                v:AddObject(handle);
            end
        end,
        CreateObject = function(self,handle)
            for i,v in pairs(self.subModules) do
                v:CreateObject(handle);
            end
        end,
        DeleteObject = function(self,handle)
            for i,v in pairs(self.subModules) do
                v:DeleteObject(handle);
            end
        end,
        Load = function(self,data)
            for i,v in pairs(self.subModules) do
                v:Load(data[i]);
            end
        end,
        Save = function(self)
            local ret = {};
            for i,v in pairs(self.subModules) do
                ret[i] = v:Save();
            end
            return ret;
        end,
        SaveDone = function(self)
            for i,v in pairs(self.subModules) do
                v:SaveDone();
            end
        end,
        CreatePlayer = function(self,id,name,team)
            for i,v in pairs(self.subModules) do
                v:CreatePlayer(id,name,team);
            end
        end,
        AddPlayer = function(self,id,name,team)
            for i,v in pairs(self.subModules) do
                v:AddPlayer(id,name,team);
            end
        end,
        DeletePlayer = function(self,id,name,team)
            for i,v in pairs(self.subModules) do
                v:DeletePlayer(id,name,team);
            end
        end,
        Receive = function(self,from,mType,...)
            for i,v in pairs(self.subModules) do
                v:Receive(from,mType,...);
            end
        end,
        Command = function(self,command,arguments)
            for i,v in pairs(self.subModules) do
                v:Command(command,arguments);
            end
        end,
        Update = function(self,dTime)
            for i,v in pairs(self.subModules) do
                v:Update(dTime);
            end
        end
    },
    mt = function(cls)
        return {__index = cls.prototype,__call = cls.new}; 
    end
}
setmetatable(mModule,mModule:mt());

local extends = function(base)
    local child = {};
    child.prototype = {super = base};

    if base then
        setmetatable(child.prototype,{__index = base.prototype});
    else
        child.prototype.new = function(self,...)
            local nInst = setmetatable({},self.mt);
            nInst:init(...);
            return nInst;
        end
    end
    --child.mt = {__index = child.prototype};
    setmetatable(child,{__index = child.prototype,__call = child.new});
    return child;
end

--simple ODF manager
--Do not save!

local odfFiles = {};
local odfHeader = {
    new = function(cls,odfFile,name)
        local inst = setmetatable({},cls:mt());
        inst.odfFile = odfFile;
        inst.header = name;
        return inst;
    end,
    prototype = {
        getVar = function(self,varName)
            return GetODFString(self.odfFile,self.header,varName,nil);
        end,
        getAsInt = function(self,varName)
            return GetODFInt(self.odfFile,self.header,varName,nil);
        end,
        getAsBool = function(self,varName)
            return GetODFBool(self.odfFile,self.header,varName,nil);
        end,
        getAsFloat = function(self,varName)
            return GetODFInt(self.odfFile,self.header,varName,nil);
        end
    },
    indexFunc = function(t,f)
        --print(bzUtils.core.odfHeader.prototype["getVar"],t,f);
        return bzUtils.core.odfHeader.prototype[f] or t:getVar(f);
    end,
    mt = function(cls)
        return {__call = cls.new,__index = cls.indexFunc};
    end
};
setmetatable(odfHeader,odfHeader:mt());
local odfFile = {
    new = function(cls,name)
        name = name or "";
        if(not odfFiles[name]) then
            local inst = setmetatable({},cls:mt());
            inst.name = name;
            inst.file = OpenODF(name);
            inst.headers = {};
            if(not inst.file) then
                error(string.format("No odf with name \"%s\"",name));
            end
            odfFiles[name] = inst;
        end
        return odfFiles[name];
    end,
    prototype = {
        getHeader = function(self,headerName)
            if(not self.headers[headerName]) then
                self.headers[headerName] = odfHeader(self.file,headerName);
            end
            return self.headers[headerName];
        end,
        de_ref = function(self)
            odfFiles[self.name] = nil;
        end
    },
    indexFunc = function(t,f)
        return bzUtils.core.odfFile.prototype[f] or t:getHeader(f);
    end,
    mt = function(cls)
        return {__call = cls.new,__index = cls.indexFunc};
    end
}

setmetatable(odfFile,odfFile:mt());

local createLocalStore = function(t,def)
    local store = setmetatable({} or def,getmetatable(t));
    return setmetatable(t,{__index = store, __newindex = store});
end

local createSaveTable = function(t)
    if(not rawget(t,"_currentSave")) then
        rawset(t,"_currentSave",{});
    end
    local mt = getmetatable(t);
    local r = rawget(t,"_currentSave");
    r["user_store"] = r["user_store"] or {};
    for i,v in pairs(mt.__index) do
        local item = v;
        if(type(v) == "table" and v.saveable and v.Save) then
            item = v:Save();
        end
        r["user_store"][i] = item;
    end
end

bzUtils = mModule:new("bzUtils");
bzUtils.storage = {};

bzUtils.storeVar = function(self,name,var)
    self.storage[name] = var;
end
bzUtils.getVar = function(self,name)
    return self.storage[name];
end



bzUtils.Save = function(self)
    local mt = getmetatable(self);
    local r = mt.__index.Save(self);
    r[0] = {
        store = {}
    };
    for i,v in pairs(self.storage) do
        local item = v;
        if(type(v) == "table" and v.saveable and v.Save) then
            item = v:Save();
        end
        r[0].store[i] = item;
    end
    self:SaveDone();
    return r;
end

bzUtils.Load = function(self,data)
    local mt = getmetatable(self);
    self.storage = data[0]["store"];
    
    mt.__index.Load(self,data);
end

bzUtils.core = {
    mModule = mModule,
    extends = extends,
    mStore = mStore,
    odfFile = odfFile,
    odfHeader = odfHeader,
    createLocalStore = createLocalStore,
    createSaveTable = createSaveTable
};



return bzUtils;